package de.tudarmstadt.langtech.substituter;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Properties;
import java.util.Vector;

import weka.classifiers.Classifier;
import weka.core.Instance;
import weka.core.Instances;
import de.tudarmstadt.langtech.substituter.settings.Configuration;
import de.tudarmstadt.langtech.substituter.util.Inventory;
import de.tudarmstadt.langtech.substituter.util.LemmaMatcher;
import de.tudarmstadt.langtech.substituter.util.TextFile;
import de.tudarmstadt.langtech.substituter.util.Util;

/**
 * Sense substituter based on a ML setup using weka
 *
 * @author cwirth
 *
 */
public class MLSenseSubstituter implements SenseSubstituter {

	// classifierMap: Map from ambiguous target to .model file name, or classifier object when already loaded
	private final HashMap<String,Object> classifierMap = new HashMap<String,Object>();
	// datasetMap: Map from ambiguous target to .ds data set file name, or data set object when already loaded
	private final HashMap<String,Object> datasetMap = new HashMap<String,Object>();
	// map from sense -> substitution list
	private HashMap<String,List<String[]>> substitutionMap = new HashMap<String,List<String[]>>();

	private Inventory[] inventories;
	// handles match for fullforms via lemma
	private LemmaMatcher lemmamatcher;
	// contains monosemous words
	private final ArrayList monoWords = new ArrayList();

	/**
	 * For API call
	 * For a list of properties, @see(de.tudarmstadt.langtech.settings.Configuraton)
	 * @param properties
	 */
	public MLSenseSubstituter(Properties properties) {
		Configuration.setConfigFile(properties);
		init();
	}

	/**
	 * For standalone call
	 * For a list of configuration parameters, @see(de.tudarmstadt.langtech.settings.Configuraton)
	 * @param properties
	 */
	public MLSenseSubstituter(String configurationFile) {
		Configuration.setConfigFile(configurationFile,false);
		init();
	}

	/**
	 * Reads the model folder and inits all configuration settings.
	 */
	protected void init() {
		File folder = new File(Configuration.modelFolder);
		for(File file : folder.listFiles()) {
			if(file.getName().endsWith(".model")) {
				String name = file.getName().substring(0,file.getName().indexOf("."));
				classifierMap.put(name, file);
			}
			if(file.getName().endsWith(".ds")) {
				String name = file.getName().substring(0,file.getName().indexOf("."));
				datasetMap.put(name, file);
			}
		}
		String [] inventoryfiles = Util.initInventories(Configuration.inventoryFile);
		substitutionMap = Util.initSubstitutions(Configuration.substitutionsFile);
		Vector ambWords = Util.readTargets(Configuration.ambiguousWordsFile);
		lemmamatcher = new LemmaMatcher(Configuration.lemmaMapFile);
		inventories=new Inventory[inventoryfiles.length];
		for(int i=0;i<inventoryfiles.length;i++) {
			inventories[i]=new Inventory(Util.inventoryEntrySuffix,Util.usePos);
			inventories[i].init(inventoryfiles[i],ambWords);
		} // rof

		TextFile mono = new TextFile(Configuration.monosemousWordsFile);
		for(String el : mono) {
			monoWords.add(el);
		}
	}

	/**
	 * returns the substitution for the word at the given index, if available
	 *
	 * @param targetIndex
	 * @param words
	 * @param postags
	 */
	public Substitution getSubstitution(int targetIndex, String[] words,String[] postags) {
		//map to lemma
		String Word = lemmamatcher.getLemma(words[targetIndex]);
		//check for monosemy
		if(monoWords.contains(Word)) {
			Substitution s = new Substitution(words[targetIndex],Word,1.0f,substitutionMap);
			return s;
		}
		//check if a classifer is existent for the given word
		if(!classifierMap.containsKey(Word)) {
			return null;
		}
		Object oc = classifierMap.get(Word);
		Classifier c;
		if(oc instanceof File) {
			c = Util.loadClassifier((File) oc);
			classifierMap.put(Word, c);
		} else {
			c = (Classifier)oc;
		}
		Object ods = datasetMap.get(Word);
		Instances ds;
		if(oc instanceof File) {
			ds = Util.loadDataset((File) ods);
			datasetMap.put(Word, ds);
		} else {
			ds = (Instances)ods;
		}
		String[] words2 = Util.makeWekaCompatible(words);
		// prepare feature reprsentation for new instance
		Vector<String> Features = Util.sentenceToFeatures(words2,postags,targetIndex,inventories);
		// classify
		String[] senseclass = classify(c, Features, ds);
		// get substitutions
		Substitution s = new Substitution(Word,senseclass[0],Float.parseFloat(senseclass[1]),substitutionMap);
		//return Substitution object
		return s;
	}

	/**
	 * Classifies the given feature representation with the given classifier
	 * Confidence is considered
	 *
	 * @param cModel
	 * @param vals
	 * @param thisDataset
	 * @return A class|confidence pair
	 */
	private String[] classify(Classifier cModel, Vector<String> vals, Instances thisDataset) {
		Instance iUse=Util.makeCompatible(vals,thisDataset);
		String maxClass="";
		double confidence=-1;
		try {
			double[] fDistribution = cModel.distributionForInstance(iUse);
			for (int i =0;i<fDistribution.length;i++) {
				if (fDistribution[i]>confidence) {
					confidence=fDistribution[i];
					maxClass=thisDataset.attribute(thisDataset.numAttributes()-1).value(i);
				}
			}
			//System.out.println("\n\nClassified:"+fDistribution[0]);
		} catch (Exception e) { System.out.println("Error with instance"+e.getMessage()+e.toString());}

		return new String[]{maxClass,Double.toString(confidence)};
	}
}
