/*
 * This is a project for adding linguistic annotations to plain text for the use of prototyping a caption highlighting system
 */

package de.tudarmstadt.langtech.substituter.training;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.ObjectOutputStream;
import java.io.OutputStreamWriter;
import java.util.Vector;

import weka.classifiers.Classifier;
import weka.core.Instances;
import weka.core.converters.ArffLoader.ArffReader;
import de.tudarmstadt.langtech.substituter.settings.Configuration;
import de.tudarmstadt.langtech.substituter.util.Inventory;
import de.tudarmstadt.langtech.substituter.util.TextFile;
import de.tudarmstadt.langtech.substituter.util.Util;
import de.tudarmstadt.langtech.viterbitagger.Viterbitagger;


/**
 *	class for training new models. Is called via run.Main
 *
 * @author cbiemann, cwirth
 */
public class TrainModels {

	static Inventory inv=new Inventory();
	static ModelPreparer mp=new ModelPreparer();
	static Viterbitagger tagger;
	static String cCall;

	/** reads 1st col of tabseparated file into vector */
	public static Vector readTargets(String filename) {
		Vector retvec=new Vector();
		TextFile instancesfile=new TextFile(filename);

		// reads in 1st col
		for(String line : instancesfile) {
			String[] cols=line.split("\t");
			if(cols[0].length()>=1) {
				retvec.add(cols[0]);
			}

		} // rof line
		return retvec;
	}

	/** SAVE MODEL */
	public void saveDataset(String arffilename, String dsfilename) {
		try {

			BufferedReader reader = new BufferedReader(new InputStreamReader(new FileInputStream(arffilename), "UTF-8"));
			ArffReader arff = new ArffReader(reader);
			Instances dataset = arff.getData();
			reader.close();
			FileOutputStream dsOutFile = new FileOutputStream(dsfilename);
			ObjectOutputStream dsOutObjectFile = new ObjectOutputStream(dsOutFile);
			dsOutObjectFile.writeObject(dataset);
			dsOutObjectFile.flush();
			dsOutObjectFile.close();

		} catch (Exception e) {System.out.println("File not found: "+e.getMessage());}
	} // end save dataset

	/**
	 * Trains and writes a classifier
	 * @param arffilename trainingsdata arff file
	 * @param modelfilename targetname for the classifer file
	 */
	public void buildSaveClassifier(String arffilename, String modelfilename) {
		try {
			BufferedReader reader = new BufferedReader(new InputStreamReader(new FileInputStream(arffilename), "UTF-8"));
			Instances data = new Instances(reader);
			reader.close();
			data.setClassIndex(data.numAttributes() - 1);
			// AODE DEFAULT
			// Classifier cModel = (Classifier) new AODE();
			//cCall = cModel.getClass().getName() + " -F 1";

			// SMO DEFAULT
			Classifier cModel = Classifier.forName(Configuration.wekaClassifier, weka.core.Utils.splitOptions(Configuration.wekaOptions));
			cCall = cModel.getClass().getName()+" "+Configuration.wekaOptions;

			// NBTree DEFAULT
			//Classifier cModel = (Classifier) new NBTree();
			//cCall = cModel.getClass().getName();

			// AODE Attrib Selected
			/*AttributeSelectedClassifier cModel = new AttributeSelectedClassifier();
			cModel.setClassifier(new AODE());
			cModel.setEvaluator(new CfsSubsetEval());
			cModel.setSearch(new BestFirst());*/

			// NBTree Attrib Selected
			/*AttributeSelectedClassifier cModel = new AttributeSelectedClassifier();
			cModel.setClassifier(new NBTree());
			cModel.setEvaluator(new CfsSubsetEval());
			cModel.setSearch(new BestFirst());*/


			cModel.buildClassifier(data);
			FileOutputStream modelOutFile = new FileOutputStream(modelfilename);
			ObjectOutputStream modelOutObjectFile = new ObjectOutputStream(modelOutFile);
			modelOutObjectFile.writeObject(cModel);
			modelOutObjectFile.flush();
			modelOutFile.close();

		} catch (Exception e) {System.out.println("File not found: "+e.getMessage());}
	} // end buildSaveAODEClassifier
	/**
	 * saves and creates the required classifer model and dataset
	 * @param arffilename
	 * @param namePrefix
	 */
	public void createModelDataSet(String arffilename, String namePrefix) {
		buildSaveClassifier(arffilename,namePrefix+".model");
		saveDataset(arffilename, namePrefix+".ds");
	}
	//* end save model */


	/** writes shell script for performing spliteval on arfffiles*/
	public static void writeSplitEval(Vector targets, String targetdir, String outfilename) {

		try {
			BufferedWriter out = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(targetdir+"/"+outfilename), "UTF-8"));


			// header
			out.write("echo Making stuff!! " +targetdir+ " \n");
			out.write("rm results*.txt\n");
			for(int i=0;i<targets.size();i++) {
				String thistarg=(String)targets.elementAt(i);
				out.write("echo Target:" +thistarg+"\n");
				out.write("java -Xmx3G -cp ../../lib/weka.jar  "+cCall+" "+Configuration.wekaOptions+" -x 10 -p 1 -t "+thistarg+".arff >> results_fold10.txt\n");
			} // rof i


			out.close();
		} catch (IOException e) {
			System.out.println("IOException:");
			e.printStackTrace();
		}



	} // end writeARFFs




	/** writes arfss for targets in targetdirs and outfile that is easy to turn into a batch steering file*/
	public void createModels(Vector targets, String instancsFilename, String targetdir, String outfilename) {
		File models = new File(targetdir);
		if(!models.exists()) {
			models.mkdirs();
		}
		for(int i=0;i<targets.size();i++) {
			String thistarg=(String)targets.elementAt(i);
			// call model preparer to write ARFF file
			mp.writeARFFfile(thistarg, instancsFilename, targetdir+"/"+thistarg+".arff") ;
			// write model and data set: weka needs prototypical classifier to memorize meta data
			createModelDataSet(targetdir+"/"+thistarg+".arff",targetdir+"/"+thistarg);

		}  // rof i
		// this is merely to create the cross validation batch file, so that quality of models can be estimated.
		// not essential for the training process
		writeSplitEval(targets,targetdir,outfilename);
	} // end makeModels



	public String[] initInventories(String inventoryListFile) {

		TextFile instancesfile=new TextFile(inventoryListFile);
		int size=0;
		int lineNr=0;
		String[] inventories=new String[0];
		for(String line : instancesfile) {

			if (lineNr==0) {
				size=(new Integer(line)).intValue(); // first line: size to be considered
				inventories=new String[size];
			} else {
				inventories[lineNr-1]=line;
			}

			lineNr+=1;

		}
		return inventories;
	} // end initInventories



	/** train class */
	public static void train() {

		TrainModels thisTrain = new TrainModels();

		// tagger
		tagger=new Viterbitagger(Configuration.taggerModelPrefix);

		// initialize
		String[] inventories=thisTrain.initInventories(Configuration.inventoryFile);

		// targs=readTargets("ambiguous_words.txt");
		Vector targs=readTargets(Configuration.ambiguousWordsFile);
		// writeARFFs(targs,"targ397-onlycluster12/","makeSplits.sh");

		//System.out.println("Using standard test example. \n targetlist, inventorylist, trainFile5col\n (ALL obligatory!)");
		System.out.println("Using arguments: \n - targetlist="+Configuration.ambiguousWordsFile+" inventorylist="+Configuration.inventoryFile+" trainFile4col="+Configuration.trainingsData+" \n");


		mp=new ModelPreparer();
		mp.init(tagger, inventories, targs);

		thisTrain.createModels(targs,Configuration.trainingsData,Configuration.modelFolder,"10-fold-CV.bat");

	} // end main




}
