/*
 * Tagger.java
 *
 *  Input format is plain text , words and interpunctuation marks seperated by spaces like in this comment .
 *  When inputting text with e.g. " e.g. " , this will be treated as one word .
 *  The tagger processes one line at the time , which is supposed to contain one sentence .
 */

package de.tudarmstadt.langtech.viterbitagger;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Collections;
import java.util.Hashtable;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.FileReader;
/**
 *
 * @author BIEMANN
 */
public class Tagger {
    boolean d=false; // debugging switch for verbose output
    boolean verbose=false; /// debuggin specific to statfram
    Lexicon lexicon;  // Contains POS lexicon
    TagList taglist;  // Contains list of tags and performs conversion Tag<->integer id
    Transitions_fast transitions_fast; 
    final int beamwidth=3;  // Viterbi beam search width. 5 is a safe choice, as 3 is usually suficient.
    final String GUESSYMBOL=""; // symbol to mark tags for OOV words
    final String SEPERATOR="/";  // symbol to seperate words from their tags
    
    
    /** Creates a new instance of Tagger */
    public Tagger(String taglistfile,String lexiconfile,String transitionfile) {
        
       
        // load taglist
        taglist=new TagList(taglistfile);        
        if (d) System.out.println("Taglist loaded:"+taglist.getNrOfTags());
        
        // load lexicon
        lexicon=new Lexicon(lexiconfile);  // lexfile, setsmooth, setpretree  
        if (d) System.out.println("Lexicon loaded");
       
        // load transitions
        transitions_fast=new Transitions_fast(transitionfile,taglist);
        if (d) System.out.println("Transitions loaded");       
    } // end public Tagger(String taglistfile,String lexiconfile,String transitionfile) constructor
    
    
    public void tagFile(String textfile) throws IOException, FileNotFoundException {
        tagFile(textfile,0);
    }

    /** tags a tokenized text string (sentence) with tagger model */
    public String tagString(String toTag) {
        String retString=new String();
        String line=toTag;
        line = "<BOS> <BOS> "+line; // add dummy words BeginOfSentence

        String[] items=line.split(" +");  // split into tokens on spaces
        int nr_of_words=items.length; // # of words in current sentence

        String[] words=new String[nr_of_words]; // SAME AS ITEMS

        for(int i=0;i<nr_of_words;i++) {
            words[i]=items[i];
        } // else fi eval


        // SAME AS ITEMS
        String[] origwords=new String[words.length];
        for(int i=0;i<words.length;i++) {
            origwords[i]=words[i];
        }

        // initialise tagmaxprob
        Vector tagmaxprob=new Vector();
        for(int i=0;i<taglist.getNrOfTags();i++) {
            double inval=0.0;
            if (i == taglist.getCodeForTag("-")) {
                inval=inval+1.0; // for upranking of BPOS tag in initialisation
            }
            tagmaxprob.add(new DoubleAndIntValue(inval, i));
        } // rof i

        // initialize maxprob and backpointer matrix

        // maxprob stores the maximal sequence probabilities for every word to have some tag
        double[][] maxprob=new double[nr_of_words][taglist.getNrOfTags()];

        // backpoint stores the tag of the last word that lead to this max prob.
        int[][] backpoint=new int[nr_of_words][taglist.getNrOfTags()];

        // initialize maxprob and backpoint for the two dummy words
        for(int i=0;i<taglist.getNrOfTags();i++) {
            if (i == taglist.getCodeForTag("-")) {
                maxprob[0][i]=1.0;
                maxprob[1][i]=1.0;
                backpoint[0][i]=taglist.getCodeForTag("-");
                backpoint[1][i]=taglist.getCodeForTag("-");
            } else {
                maxprob[0][i]=0;
                maxprob[1][i]=0;
                backpoint[0][i]=taglist.getCodeForTag("-");
                backpoint[1][i]=taglist.getCodeForTag("-");
            } // esle fi.
        } // rof i

        if (d) System.out.println("Initialized matrices for "+nr_of_words+" words and "+taglist.getNrOfTags()+" tags");

        // loop over words and fill matrices
        for(int pos=2;pos<nr_of_words;pos++) {

            // obtain probabilities per tag for current word
            double[] allLexProbs=lexicon.getTagDistribution(words[pos]);
            Collections.sort(tagmaxprob);
            Vector acceptTags=new Vector();
            int count=0;

            double absmaxprob=0.0; //for multiplication in case of underflow
            for(Enumeration e=tagmaxprob.elements();e.hasMoreElements();) {

                DoubleAndIntValue current=(DoubleAndIntValue)e.nextElement();
                if (count<beamwidth) {
                    count++;
                    if (current.getProbability()>0.0) {
                        acceptTags.add(new Integer(current.getTagNr()));
                        if (d) System.out.println("accepted: tagnr "+current.getTagNr()+" with p="+current.getProbability());
                        if (absmaxprob<current.getProbability()) {absmaxprob=current.getProbability();}
                    } // fi prob>0
                } // fi beamwidth
            } // rof

            // underflow management: multiply with 1E100 if maxprob<1E-200. This is a machine precision hack
            if (absmaxprob<1E-200) {
                for(int i=0;i<taglist.getNrOfTags();i++) {
                    maxprob[pos-1][i]=maxprob[pos-1][i]*1E100;
                } // rof
            } // fi underflow

            tagmaxprob=new Vector();

            for(int tag=0;tag<taglist.getNrOfTags();tag++) {

                if (d) System.out.println("Processing tag for word "+tag+" "+words[pos]);

                /* for current tag, compute all probabilities
                 *   p(tag|T1,T2) =                sequence probability fot tag taking way T1 T2 =
                 *      maxprob[pos-1][T2] *            sequence probability ending on T2 *
                 *      transition(tag|T1,T2) *         probability of tag given last 2 tags *
                 *      lex(tag|word).                  probability of tag, given word
                 *
                 *   This is iterated over all possible T2s. The maximal prob and the respective T2 is saved in matrices
                */

                double curr_maxprob=-1.0;
                int curr_backpoint=-1;
                double lexprob=allLexProbs[tag];


                for(Enumeration pt=acceptTags.elements();pt.hasMoreElements();) {

                    int prevtag=((Integer)pt.nextElement()).intValue();
                    int prevprevtag=backpoint[pos-1][prevtag];  // code of T1
                    double prev_sequence_prob=maxprob[pos-1][prevtag];  // maxprob of T2
                    double trans_prob;
                    trans_prob=transitions_fast.getTransProb(prevprevtag, prevtag, tag);
                    double curr_prob=prev_sequence_prob*trans_prob*lexprob;

                    if ((d)&&(curr_prob>0.0)) System.out.println("word: "+words[pos]+ "   Tag seq: "+prevprevtag + " "+prevtag +" "+tag+"  prob: "+curr_prob);

                    if (curr_prob>curr_maxprob) {
                        curr_maxprob=curr_prob;
                        curr_backpoint=prevtag;
                    } // fi. curr>max
                } // rof prevtag

                // save in matrices
                maxprob[pos][tag]=curr_maxprob;
                backpoint[pos][tag]=curr_backpoint;
                if (d) System.out.println("Tag and backpointer: "+tag+" "+curr_backpoint);

                tagmaxprob.add(new DoubleAndIntValue(curr_maxprob,tag));

            } // rof tag

        } // rof pos

        // all max paths have been determined. Chose and output tag sequence
        String tagSequence[]= new String[nr_of_words];

        int maxtag=0;
        double maxtag_prob=-1;
        for(int tag=0;tag<taglist.getNrOfTags();tag++) {
            if (maxprob[nr_of_words-1][tag]>maxtag_prob) {
                maxtag=tag;
                maxtag_prob=maxprob[nr_of_words-1][tag];
            } //fi. maxprob
        } // rof tag

        for(int pos=nr_of_words-1;pos>=0;pos--) {
            tagSequence[pos]=taglist.getTagForCode(maxtag);
            maxtag=backpoint[pos][maxtag];
        } // rof pos

        if (d) System.out.println(" ----- Sentence result: ------- ");

        // pos starts with 2 to omit begin-of-sentence words; put back in vector
        for(int pos=2;pos<nr_of_words;pos++) {
             String guess="";
             boolean inLex=lexicon.containsWord(words[pos]);
             if (!inLex) {guess=GUESSYMBOL;}


             // OUTPUT of words
             retString+=" "+origwords[pos]+SEPERATOR+tagSequence[pos]+guess;
        } // rof pos



        if (verbose) System.out.println(retString);




        return retString;
    } // end tagString

    /** tags a statfram vector with tagger model*/
    public Vector tagVector(Vector toTag) {
        // NO POS for 1st token

        String sentence="";
        int startIndex=0;
        Hashtable prev=new Hashtable();
        prev.put("breakAfter","1");
        Hashtable lasttok=new Hashtable();
        lasttok.put("word","EOF");
        lasttok.put("breakBefore","1000");
        toTag.add(lasttok);

        boolean breakpoint=false;

        for(int vpos=0;vpos<toTag.size();vpos++) {
           Hashtable curr=(Hashtable)toTag.elementAt(vpos);
           breakpoint=false;
           if (prev.containsKey("breakAfter")||curr.containsKey("breakBefore")) {breakpoint=true;}
           if ((!breakpoint)) {sentence=sentence+" "+(String)curr.get("word");}
           else {  // TAG IT
                String line=sentence;
                line = "<BOS> <BOS> "+line; // add dummy words BeginOfSentence

                String[] items=line.split(" +");  // split into tokens on spaces
                int nr_of_words=items.length; // # of words in current sentence

                String[] words=new String[nr_of_words]; // SAME AS ITEMS

                for(int i=0;i<nr_of_words;i++) {
                    words[i]=items[i];
                } // else fi eval


                // SAME AS ITEMS
                String[] origwords=new String[words.length];
                for(int i=0;i<words.length;i++) {
                    origwords[i]=words[i];
                }

                // initialise tagmaxprob
                Vector tagmaxprob=new Vector();
                for(int i=0;i<taglist.getNrOfTags();i++) {
                    double inval=0.0;
                    if (i == taglist.getCodeForTag("-")) {
                        inval=inval+1.0; // for upranking of BPOS tag in initialisation
                    }
                    tagmaxprob.add(new DoubleAndIntValue(inval, i));
                } // rof i

                // initialize maxprob and backpointer matrix

                // maxprob stores the maximal sequence probabilities for every word to have some tag
                double[][] maxprob=new double[nr_of_words][taglist.getNrOfTags()];

                // backpoint stores the tag of the last word that lead to this max prob.
                int[][] backpoint=new int[nr_of_words][taglist.getNrOfTags()];

                // initialize maxprob and backpoint for the two dummy words
                for(int i=0;i<taglist.getNrOfTags();i++) {
                    if (i == taglist.getCodeForTag("-")) {
                        maxprob[0][i]=1.0;
                        maxprob[1][i]=1.0;
                        backpoint[0][i]=taglist.getCodeForTag("-");
                        backpoint[1][i]=taglist.getCodeForTag("-");
                    } else {
                        maxprob[0][i]=0;
                        maxprob[1][i]=0;
                        backpoint[0][i]=taglist.getCodeForTag("-");
                        backpoint[1][i]=taglist.getCodeForTag("-");
                    } // esle fi.
                } // rof i

                if (d) System.out.println("Initialized matrices for "+nr_of_words+" words and "+taglist.getNrOfTags()+" tags");

                // loop over words and fill matrices
                for(int pos=2;pos<nr_of_words;pos++) {

                    // obtain probabilities per tag for current word
                    double[] allLexProbs=lexicon.getTagDistribution(words[pos]);
                    Collections.sort(tagmaxprob);
                    Vector acceptTags=new Vector();
                    int count=0;

                    double absmaxprob=0.0; //for multiplication in case of underflow
                    for(Enumeration e=tagmaxprob.elements();e.hasMoreElements();) {

                        DoubleAndIntValue current=(DoubleAndIntValue)e.nextElement();
                        if (count<beamwidth) {
                            count++;
                            if (current.getProbability()>0.0) {
                                acceptTags.add(new Integer(current.getTagNr()));
                                if (d) System.out.println("accepted: tagnr "+current.getTagNr()+" with p="+current.getProbability());
                                if (absmaxprob<current.getProbability()) {absmaxprob=current.getProbability();}
                            } // fi prob>0
                        } // fi beamwidth
                    } // rof

                    // underflow management: multiply with 1E100 if maxprob<1E-200. This is a machine precision hack
                    if (absmaxprob<1E-200) {
                        for(int i=0;i<taglist.getNrOfTags();i++) {
                            maxprob[pos-1][i]=maxprob[pos-1][i]*1E100;
                        } // rof
                    } // fi underflow

                    tagmaxprob=new Vector();

                    for(int tag=0;tag<taglist.getNrOfTags();tag++) {

                        if (d) System.out.println("Processing tag for word "+tag+" "+words[pos]);

                        /* for current tag, compute all probabilities
                         *   p(tag|T1,T2) =                sequence probability fot tag taking way T1 T2 =
                         *      maxprob[pos-1][T2] *            sequence probability ending on T2 *
                         *      transition(tag|T1,T2) *         probability of tag given last 2 tags *
                         *      lex(tag|word).                  probability of tag, given word
                         *
                         *   This is iterated over all possible T2s. The maximal prob and the respective T2 is saved in matrices
                        */

                        double curr_maxprob=-1.0;
                        int curr_backpoint=-1;
                        double lexprob=allLexProbs[tag];


                        for(Enumeration pt=acceptTags.elements();pt.hasMoreElements();) {

                            int prevtag=((Integer)pt.nextElement()).intValue();
                            int prevprevtag=backpoint[pos-1][prevtag];  // code of T1
                            double prev_sequence_prob=maxprob[pos-1][prevtag];  // maxprob of T2
                            double trans_prob;
                            trans_prob=transitions_fast.getTransProb(prevprevtag, prevtag, tag);
                            double curr_prob=prev_sequence_prob*trans_prob*lexprob;

                            if ((d)&&(curr_prob>0.0)) System.out.println("word: "+words[pos]+ "   Tag seq: "+prevprevtag + " "+prevtag +" "+tag+"  prob: "+curr_prob);

                            if (curr_prob>curr_maxprob) {
                                curr_maxprob=curr_prob;
                                curr_backpoint=prevtag;
                            } // fi. curr>max
                        } // rof prevtag

                        // save in matrices
                        maxprob[pos][tag]=curr_maxprob;
                        backpoint[pos][tag]=curr_backpoint;
                        if (d) System.out.println("Tag and backpointer: "+tag+" "+curr_backpoint);

                        tagmaxprob.add(new DoubleAndIntValue(curr_maxprob,tag));

                    } // rof tag

                } // rof pos

                // all max paths have been determined. Chose and output tag sequence
                String tagSequence[]= new String[nr_of_words];

                int maxtag=0;
                double maxtag_prob=-1;
                for(int tag=0;tag<taglist.getNrOfTags();tag++) {
                    if (maxprob[nr_of_words-1][tag]>maxtag_prob) {
                        maxtag=tag;
                        maxtag_prob=maxprob[nr_of_words-1][tag];
                    } //fi. maxprob
                } // rof tag

                for(int pos=nr_of_words-1;pos>=0;pos--) {
                    tagSequence[pos]=taglist.getTagForCode(maxtag);
                    maxtag=backpoint[pos][maxtag];
                } // rof pos

                if (d) System.out.println(" ----- Sentence result: ------- ");

                // pos starts with 2 to omit begin-of-sentence words; put back in vector
                for(int pos=2;pos<nr_of_words;pos++) {
                     String guess="";
                     boolean inLex=lexicon.containsWord(words[pos]);
                     if (!inLex) {guess=GUESSYMBOL;}
                     ((Hashtable)toTag.elementAt(startIndex-3+pos)).put("POS",tagSequence[pos]);

                     // OUTPUT of words
                     if (verbose) System.out.print(" "+origwords[pos]+SEPERATOR+tagSequence[pos]+guess);
                } // rof pos



              if (verbose) System.out.println();


                sentence=(String)curr.get("word");
                startIndex=vpos+1;
           }

           prev=curr;
        }

        // remove last dummy
        toTag.removeElementAt(toTag.size()-1);

        return toTag;
    }


    /** tags a text file with tagger model */
    public void tagFile(String textfile, int column) throws IOException, FileNotFoundException{
        
        BufferedReader textFileReader = new BufferedReader(new FileReader(textfile)); // file access 
        String line;    // the current line 
        Vector tagmaxprob=new Vector(); // for beam search
        String[] origwords;  // tokens from line
        
        // loop over lines=sentences
        while((line=textFileReader.readLine())!=null) {
            if (d) System.out.println("Tagging line:"+line);
            String[] lineparts=line.split("\t");
            for (int i=0;i<column;i++) {
                System.out.print(lineparts[i]+"\t");
            }
            line=lineparts[column];

            line = "<BOS> <BOS> "+line; // add dummy words BeginOfSentence
                        
            String[] items=line.split(" +");  // split into tokens on spaces
            int nr_of_words=items.length; // # of words in current sentence
            
            String[] words=new String[nr_of_words]; // SAME AS ITEMS
            String[] gold_tags=new String[nr_of_words];
            
            for(int i=0;i<nr_of_words;i++) {
                words[i]=items[i];
            } // else fi eval 
                         
                
            // SAME AS ITEMS
            origwords=new String[words.length];
            for(int i=0;i<words.length;i++) {
                origwords[i]=words[i];   
            }
            
            // initialise tagmaxprob
            tagmaxprob=new Vector();
            for(int i=0;i<taglist.getNrOfTags();i++) {
                double inval=0.0;          
                if (i == taglist.getCodeForTag("-")) {
                    inval=inval+1.0; // for upranking of BPOS tag in initialisation                  
                }
                tagmaxprob.add(new DoubleAndIntValue(inval, i));
            } // rof i
            
            // initialize maxprob and backpointer matrix
            
            // maxprob stores the maximal sequence probabilities for every word to have some tag
            double[][] maxprob=new double[nr_of_words][taglist.getNrOfTags()];
            
            // backpoint stores the tag of the last word that lead to this max prob.
            int[][] backpoint=new int[nr_of_words][taglist.getNrOfTags()];
            
            // initialize maxprob and backpoint for the two dummy words
            for(int i=0;i<taglist.getNrOfTags();i++) {         
                if (i == taglist.getCodeForTag("-")) {
                    maxprob[0][i]=1.0;
                    maxprob[1][i]=1.0;
                    backpoint[0][i]=taglist.getCodeForTag("-");
                    backpoint[1][i]=taglist.getCodeForTag("-");    
                } else {
                    maxprob[0][i]=0;
                    maxprob[1][i]=0;
                    backpoint[0][i]=taglist.getCodeForTag("-");
                    backpoint[1][i]=taglist.getCodeForTag("-");
                } // esle fi.
            } // rof i
            
            if (d) System.out.println("Initialized matrices for "+nr_of_words+" words and "+taglist.getNrOfTags()+" tags");
            
            // loop over words and fill matrices
            for(int pos=2;pos<nr_of_words;pos++) {
                
                // obtain probabilities per tag for current word
                double[] allLexProbs=lexicon.getTagDistribution(words[pos]);          
                Collections.sort(tagmaxprob);
                Vector acceptTags=new Vector();
                int count=0;
                
                double absmaxprob=0.0; //for multiplication in case of underflow
                for(Enumeration e=tagmaxprob.elements();e.hasMoreElements();) {
                    
                    DoubleAndIntValue current=(DoubleAndIntValue)e.nextElement();
                    if (count<beamwidth) {
                        count++;
                        if (current.getProbability()>0.0) {
                            acceptTags.add(new Integer(current.getTagNr()));
                            if (d) System.out.println("accepted: tagnr "+current.getTagNr()+" with p="+current.getProbability());
                            if (absmaxprob<current.getProbability()) {absmaxprob=current.getProbability();}
                        } // fi prob>0
                    } // fi beamwidth
                } // rof
                
                // underflow management: multiply with 1E100 if maxprob<1E-200. This is a machine precision hack
                if (absmaxprob<1E-200) {
                    for(int i=0;i<taglist.getNrOfTags();i++) {
                        maxprob[pos-1][i]=maxprob[pos-1][i]*1E100;
                    } // rof
                } // fi underflow
                              
                tagmaxprob=new Vector();
                
                for(int tag=0;tag<taglist.getNrOfTags();tag++) {
                    
                    if (d) System.out.println("Processing tag for word "+tag+" "+words[pos]);
                    
                    /* for current tag, compute all probabilities
                     *   p(tag|T1,T2) =                sequence probability fot tag taking way T1 T2 =
                     *      maxprob[pos-1][T2] *            sequence probability ending on T2 *
                     *      transition(tag|T1,T2) *         probability of tag given last 2 tags *
                     *      lex(tag|word).                  probability of tag, given word
                     *
                     *   This is iterated over all possible T2s. The maximal prob and the respective T2 is saved in matrices
                    */   
                    
                    double curr_maxprob=-1.0;
                    int curr_backpoint=-1;
                    double lexprob=allLexProbs[tag];
                    
                                                          
                    for(Enumeration pt=acceptTags.elements();pt.hasMoreElements();) {    
                       
                        int prevtag=((Integer)pt.nextElement()).intValue();
                        int prevprevtag=backpoint[pos-1][prevtag];  // code of T1
                        double prev_sequence_prob=maxprob[pos-1][prevtag];  // maxprob of T2
                        double trans_prob; 
                        trans_prob=transitions_fast.getTransProb(prevprevtag, prevtag, tag);                        
                        double curr_prob=prev_sequence_prob*trans_prob*lexprob;                       
                        
                        if ((d)&&(curr_prob>0.0)) System.out.println("word: "+words[pos]+ "   Tag seq: "+prevprevtag + " "+prevtag +" "+tag+"  prob: "+curr_prob);                 
                        
                        if (curr_prob>curr_maxprob) {
                            curr_maxprob=curr_prob;
                            curr_backpoint=prevtag;
                        } // fi. curr>max
                    } // rof prevtag
                    
                    // save in matrices
                    maxprob[pos][tag]=curr_maxprob;
                    backpoint[pos][tag]=curr_backpoint;                   
                    if (d) System.out.println("Tag and backpointer: "+tag+" "+curr_backpoint);
                    
                    tagmaxprob.add(new DoubleAndIntValue(curr_maxprob,tag));
                    
                } // rof tag
          
            } // rof pos
            
            // all max paths have been determined. Chose and output tag sequence
            String tagSequence[]= new String[nr_of_words];
            
            int maxtag=0;
            double maxtag_prob=-1;
            for(int tag=0;tag<taglist.getNrOfTags();tag++) {
                if (maxprob[nr_of_words-1][tag]>maxtag_prob) {
                    maxtag=tag;
                    maxtag_prob=maxprob[nr_of_words-1][tag];
                } //fi. maxprob
            } // rof tag
            
            for(int pos=nr_of_words-1;pos>=0;pos--) {
                tagSequence[pos]=taglist.getTagForCode(maxtag);
                maxtag=backpoint[pos][maxtag];
            } // rof pos
            
            if (d) System.out.println(" ----- Sentence result: ------- ");
                
            // pos starts with 2 to omit begin-of-sentence words
            for(int pos=2;pos<nr_of_words;pos++) { 
                 String guess="";
                 boolean inLex=lexicon.containsWord(words[pos]);              
                 if (!inLex) {guess=GUESSYMBOL;} 
                 
                 // OUTPUT of words              
                 System.out.print(" "+origwords[pos]+SEPERATOR+tagSequence[pos]+guess); 
            } // rof pos
            
            for (int i=column+1;i<lineparts.length;i++) {
                System.out.print("\t"+lineparts[i]);
            }
            
            // OUTPUT: End of line 
            System.out.println();
            
        } // elihw loop over sentences
        
        // file access end    
        textFileReader.close(); 
      
    } // end void tagFile();
     
} // end public class Tagger


// for comparing and sorting tag numbers with probabilities, necessary for beam search
class DoubleAndIntValue implements Comparable{
    private double probability;
    private int tagnumber;
    
    /**
     * Creates new Object.

     */
    public DoubleAndIntValue(double _prob,int _tag){
        this.probability=_prob;
        this.tagnumber=_tag;
    }
    
    public int compareTo(Object o){
        if(((DoubleAndIntValue) o).getProbability() > probability)return 1;
        else if(((DoubleAndIntValue) o).getProbability() < probability)return -1;
        else return 0;
    }

    /**
     * Get the tagnr
     * @return tag number.
     */
    public int getTagNr(){
        return this.tagnumber;
    }
    /**
     * Get the new word number.
     * @return The old word number.
     */
    public double getProbability(){
        return this.probability;
    }

}
